@@ -1,31 +1,64 @@
+# snoo.py
 import asyncio
 import json
 import logging
 import secrets
 import uuid
 from datetime import datetime as dt
-from typing import Callable
 
 import aiohttp
-import aiomqtt
-from pubnub.enums import PNReconnectionPolicy
+from containers import AuthorizationInfo, SnooData, SnooDevice, SnooStates
+from exceptions import InvalidSnooAuth, SnooAuthException, SnooDeviceError
+from pubnub.callbacks import SubscribeCallback
+from pubnub.enums import PNReconnectionPolicy, PNStatusCategory
 from pubnub.pnconfiguration import PNConfiguration
 from pubnub.pubnub_asyncio import PubNubAsyncio
+from pubnub_async import SnooPubNub
+
+
+class MySubscribeCallback(SubscribeCallback):
+    def __init__(self, snoo_instance, device_id, message_type):
+        self.snoo_instance = snoo_instance
+        self.device_id = device_id
+        self.message_type = message_type
+
+    def status(self, pubnub, status):
+        if status.category == PNStatusCategory.PNUnexpectedDisconnectCategory:
+            pass  # Handle unexpected disconnect
+        elif status.category == PNStatusCategory.PNConnectedCategory:
+            pass  # This is expected for a subscribe, this means we are connected and subscribing
+        elif status.category == PNStatusCategory.PNReconnectedCategory:
+            pass  # Happens as part of a reconnect process
+        elif status.category == PNStatusCategory.PNDisconnectedCategory:
+            pass
+        elif status.is_error():
+            _LOGGER.error(f"Subscription error: {status}")
+
+    def presence(self, pubnub, presence):
+        pass  # Handle presence events if needed
+
+    def message(self, pubnub, message):
+        if self.message_type == "Activity":
+            if "system_state" in message.message:
+                try:
+                    data = SnooData.from_dict(message.message)
+                    _LOGGER.debug(data)
+                    self.snoo_instance.data_map[self.device_id] = data
+                    if self.device_id in self.snoo_instance.subscription_functions:
+                        self.snoo_instance.subscription_functions[self.device_id]()
+                except Exception as e:
+                    _LOGGER.error(f"Error processing message: {e}, Message: {message.message}")
+            else:
+                _LOGGER.debug(f"[{self.device_id}] Got message on {self.message_type} of {message.message}")
+        else:
+            _LOGGER.debug(f"[{self.device_id}] Got message on {self.message_type} of {message.message}")
 
-from .containers import (
-    AuthorizationInfo,
-    SnooData,
-    SnooDevice,
-    SnooStates,
-)
-from .exceptions import InvalidSnooAuth, SnooAuthException, SnooCommandException, SnooDeviceError
-from .pubnub_async import SnooPubNub
 
 _LOGGER = logging.getLogger(__name__)
 
 
 class Snoo:
-    def __init__(self, email: str, password: str, clientsession: aiohttp.ClientSession):
+    def __init__(self, email, password, clientsession: aiohttp.ClientSession):
         self.email = email
         self.password = password
         self.session = clientsession
@@ -58,7 +91,6 @@ class Snoo:
             "user-agent": "okhttp/4.12.0",
             "accept": "application/json",
         }
-
         self.aws_auth_data = {
             "AuthParameters": {
                 "PASSWORD": self.password,
@@ -88,35 +120,34 @@ class Snoo:
         self.subscription_functions = {}
         self.data_map = {}
         self.pubnub_instances: dict[str, SnooPubNub] = {}
-        self.reauth_task: asyncio.Task | None = None
-        self._client_map: dict[str, aiomqtt.Client] = {}
-        self._mqtt_tasks: dict[str, asyncio.Task] = {}
-        self._client_cond = asyncio.Condition()
 
-    async def refresh_tokens(self):
-        # TODO: Figure out hwo to get this to work and not do a serializaiton exception
-        data = {
-            "AuthParameters": {"REFRESH_TOKEN": self.tokens.aws_refresh},
-            "AuthFlow": "REFRESH_TOKEN_AUTH",
-            "ClientId": "6kqofhc8hm394ielqdkvli0oea",
-        }
-        r = await self.session.post(self.aws_auth_url, data=data, headers=self.aws_auth_hdr)
-        resp = await r.json(content_type=None)
-        if "__type" in resp and resp["__type"] == "NotAuthorizedException":
-            raise InvalidSnooAuth()
-        result = resp["AuthenticationResult"]
-        self.tokens = AuthorizationInfo(aws_id=result[""])
+    # async def refresh_tokens(self):
+    #     # TODO: Figure out hwo to get this to work and not do a serializaiton exception
+    #     data = {
+    #         "AuthParameters": {
+    #             "REFRESH_TOKEN": self.tokens.aws_refresh
+    #         },
+    #         "AuthFlow": "REFRESH_TOKEN_AUTH",
+    #         "ClientId": "6kqofhc8hm394ielqdkvli0oea"
+    #
+    #     }
+    #     r = await self.session.post(self.aws_auth_url, data=data, headers=self.aws_auth_hdr)
+    #     resp = await r.json(content_type=None)
+    #     if "__type" in resp and  resp["__type"] == "NotAuthorizedException":
+    #         raise InvalidSnooAuth()
+    #     result = resp['AuthenticationResult']
+    #     self.tokens = AuthorizationInfo(aws_id=result[""])
 
     def check_tokens(self):
         if self.tokens is None:
             raise Exception("You need to authenticate before you continue")
 
-    def generate_snoo_auth_headers(self, amz_token: str) -> dict[str, str]:
+    def generate_snoo_auth_headers(self, amz_token):
         hdrs = self.snoo_auth_hdr.copy()
         hdrs["authorization"] = f"Bearer {amz_token}"
         return hdrs
 
-    def generate_snoo_data_url(self, device_id: str | float, snoo_token: str) -> str:
+    def generate_snoo_data_url(self, device_id, snoo_token):
         if isinstance(device_id, float):
             device_id = str(int(device_id))
         req_uuid = uuid.uuid1()
@@ -127,24 +158,24 @@ class Snoo:
         url = f"https://happiestbaby.pubnubapi.com/v2/history/sub-key/sub-c-97bade2a-483d-11e6-8b3b-02ee2ddab7fe/channel/ActivityState.{device_id}?pnsdk=PubNub-Kotlin%2F7.4.0&l_pub=0.064&auth={snoo_token}&requestid={req_uuid}&include_token=true&count=1&include_meta=false&reverse=false&uuid=android_{app_dev_id}_{dev_uuid}"
         return url
 
-    def generate_id(self) -> str:
+    def generate_id(self):
         app_dev_id_len = 24
         n = app_dev_id_len * 3 // 4
         app_dev_id = secrets.token_urlsafe(n)
         return app_dev_id
 
-    async def subscribe(self, device: SnooDevice, function: Callable):
+    async def subscribe(self, device: SnooDevice, function):
         pnconfig = PNConfiguration()
         pnconfig.subscribe_key = "sub-c-97bade2a-483d-11e6-8b3b-02ee2ddab7fe"
         pnconfig.publish_key = "pub-c-699074b0-7664-4be2-abf8-dcbb9b6cd2bf"
         pnconfig.user_id = secrets.token_urlsafe(16)
         pnconfig.auth_key = self.tokens.snoo
         pnconfig.reconnect_policy = PNReconnectionPolicy.EXPONENTIAL
-        self.pubnub = PubNubAsyncio(pnconfig)
+        pubnub = PubNubAsyncio(pnconfig)
         device_id = device.serialNumber
 
         if device_id not in self.pubnub_instances:
-            self.pubnub_instances[device_id] = SnooPubNub(self.pubnub, device_id)
+            self.pubnub_instances[device_id] = SnooPubNub(pubnub, device_id)
         pubnub_instance = self.pubnub_instances[device_id]
         unsub = pubnub_instance.subscribe(function)
         asyncio.create_task(pubnub_instance.run())
@@ -160,68 +191,32 @@ class Snoo:
                     pass
         self.pubnub_instances = {}
 
-        for task in self._mqtt_tasks.values():
-            task.cancel()
-        await asyncio.gather(*self._mqtt_tasks.values(), return_exceptions=True)
-        self._mqtt_tasks = {}
-
-        if self.reauth_task:
-            self.reauth_task.cancel()
-            self.reauth_task = None
-
     def publish_callback(self, result, status):
         if status.is_error():
             _LOGGER.warning(f"Message failed with {status.status_code}, {status.error_data.__dict__}")
 
-    async def send_command(self, command: str, device: SnooDevice, **kwargs):
+    def send_command(self, command: str, device: SnooDevice, **kwargs):
         ts = int(dt.now().timestamp() * 10_000_000)
-        try:
-            # Acquire the condition lock
-            async with self._client_cond:
-                try:
-                    # Wait up to 30 seconds for the client to connect.
-                    # The wait_for() method will wait until the lambda returns True.
-                    # It releases the lock while waiting and re-acquires it before returning.
-                    await asyncio.wait_for(
-                        self._client_cond.wait_for(lambda: device.serialNumber in self._client_map), timeout=30.0
-                    )
-                except asyncio.TimeoutError:
-                    _LOGGER.error(f"Timed out waiting for client for device {device.serialNumber} to connect.")
-                    raise SnooCommandException(f"Client for device {device.serialNumber} is not connected.") from None
+        self.pubnub.publish().channel(f"ControlCommand.{device.serialNumber}").message(
+            {"ts": ts, "command": command, **kwargs}
+        ).pn_async(self.publish_callback)
 
-                # Once wait_for returns, we know the client exists and we hold the lock.
-                await self._client_map[device.serialNumber].publish(
-                    topic=f"{device.awsIoT.thingName}/state_machine/control",
-                    payload=json.dumps({"ts": ts, "command": command, **kwargs}),
-                )
-        except Exception as e:
-            raise SnooCommandException from e
+    def start_snoo(self, device: SnooDevice):
+        self.send_command("start_snoo", device)
 
-    async def start_snoo(self, device: SnooDevice):
-        await self.send_command("start_snoo", device)
+    def stop_snoo(self, device: SnooDevice):
+        self.send_command("go_to_state", device, **{"state": "ONLINE", "hold": "off"})
 
-    async def stop_snoo(self, device: SnooDevice):
-        await self.send_command("go_to_state", device, **{"state": "ONLINE", "hold": "off"})
+    def set_level(self, device: SnooDevice, level: SnooStates):
+        self.send_command("go_to_state", device, **{"state": level.value, "hold": "off"})
 
-    async def set_level(self, device: SnooDevice, level: SnooStates, hold: bool = False):
-        if hold:
-            hold = "on"
-        else:
-            hold = "off"
+    def set_sticky_white_noise(self, device: SnooDevice, on: bool):
+        self.send_command("set_sticky_white_noise", device, **{"state": "on" if on else "off", "timeout_min": 15})
 
-        await self.send_command("go_to_state", device, **{"state": level.value, "hold": hold})
+    def get_status(self, device: SnooDevice):
+        self.send_command("send_status", device)
 
-    async def set_sticky_white_noise(self, device: SnooDevice, on: bool):
-        await self.send_command(
-            "set_sticky_white_noise",
-            device,
-            **{"state": "on" if on else "off", "timeout_min": 15},
-        )
-
-    async def get_status(self, device: SnooDevice):
-        await self.send_command("send_status", device)
-
-    async def auth_amazon(self) -> dict:
+    async def auth_amazon(self):
         r = await self.session.post(self.aws_auth_url, data=self.aws_auth_data, headers=self.aws_auth_hdr)
         resp = await r.json(content_type=None)
         if "__type" in resp and resp["__type"] == "NotAuthorizedException":
@@ -229,10 +224,10 @@ class Snoo:
         result = resp["AuthenticationResult"]
         return result
 
-    async def auth_snoo(self, id_token: str) -> dict:
+    async def auth_snoo(self, id_token):
         hdrs = self.generate_snoo_auth_headers(id_token)
         r = await self.session.post(self.snoo_auth_url, data=self.snoo_auth_data, headers=hdrs)
-        return await r.json()
+        return r
 
     async def authorize(self) -> AuthorizationInfo:
         try:
@@ -240,28 +235,27 @@ class Snoo:
             access = amz["AccessToken"]
             _id = amz["IdToken"]
             ref = amz["RefreshToken"]
-            snoo_token = await self.auth_snoo(_id)
+            snoo = await self.auth_snoo(_id)
+            snoo_token = await snoo.json()
             snoo_expiry = snoo_token["expiresIn"] / 1.5
             snoo_token = snoo_token["snoo"]["token"]
             self.tokens = AuthorizationInfo(snoo=snoo_token, aws_access=access, aws_id=_id, aws_refresh=ref)
-            self.reauth_task = asyncio.create_task(self.schedule_reauthorization(snoo_expiry))
+            asyncio.create_task(self.schedule_reauthorization(snoo_expiry))
         except InvalidSnooAuth as ex:
             raise ex
         except Exception as ex:
             raise SnooAuthException from ex
         return self.tokens
 
-    async def schedule_reauthorization(self, snoo_expiry: float):
-        _LOGGER.info("Snoo token has expired - reauthorizing...")
+    async def schedule_reauthorization(self, snoo_expiry: int):
         try:
             await asyncio.sleep(snoo_expiry)
             await self.authorize()
             for instance in self.pubnub_instances.values():
                 instance.update_token(self.tokens.snoo)
-            self.pubnub.config.auth_token = self.tokens.snoo
 
         except Exception as ex:
-            _LOGGER.exception(f"Error during reauthorization: {ex}")
+            print(f"Error during reauthorization: {ex}")
 
     async def get_devices(self) -> list[SnooDevice]:
         hdrs = self.generate_snoo_auth_headers(self.tokens.aws_id)
@@ -273,61 +267,27 @@ class Snoo:
         devs = [SnooDevice.from_dict(dev) for dev in resp["snoo"]]
         return devs
 
-    def start_subscribe(self, device: SnooDevice, function: Callable):
-        if device.serialNumber in self._mqtt_tasks and not self._mqtt_tasks[device.serialNumber].done():
-            _LOGGER.warning(f"Subscription task for device {device.serialNumber} is already running.")
-            return
+    # async def set_volume_level(self, level, baby_id: str):
+    # f = await self.session.patch(
+    #     self.snoo_baby_url + baby_id,
+    #     headers={"authorization": f"Bearer {self.tokens.aws_id}"},
+    #     json={"settings": {"minimalLevelVolume": level}},
+    # )
 
-        self._mqtt_tasks[device.serialNumber] = asyncio.create_task(self.subscribe_mqtt(device, function))
+    # lowest = 'lvl-2'
+    # low='lvl-1'
+    # normal='lvl10'
+    # high='lvl+1'
+    # highest='lvl+2'
+    #
+    #
+    # soothing
+    # normal='lvl10'
+    # high='lvl+1'
+    # highest='lvl+2'
 
-    async def subscribe_mqtt(self, device: SnooDevice, function: Callable):
-        host = device.awsIoT.clientEndpoint
-        port = 443
-        websocket_path = "/mqtt"
-        token = self.tokens.aws_id
-        headers = {"token": token}
-        password = None
-        client_id = f"HA_{uuid.uuid4()}"
-        user_name = "?SDK=iOS&Version=2.40.1"
 
-        logging.debug(f"Attempting to connect to wss://{host}:{port}{websocket_path}")
-
-        try:
-            async with aiomqtt.Client(
-                hostname=host,
-                port=port,
-                username=user_name,
-                password=password,
-                identifier=client_id,
-                transport="websockets",
-                websocket_path=websocket_path,
-                websocket_headers=headers,
-                tls_params=aiomqtt.TLSParameters(),
-                protocol=aiomqtt.ProtocolVersion.V31,
-                timeout=10,
-            ) as client:
-                logging.info(f"✅ Successfully connected to MQTT broker for {device.serialNumber}!")
-
-                # Acquire the lock, add the client to the map, and notify waiting tasks.
-                async with self._client_cond:
-                    self._client_map[device.serialNumber] = client
-                    self._client_cond.notify_all()
-
-                topic = f"{device.awsIoT.thingName}/state_machine/activity_state"
-                await client.subscribe(topic)
-                logging.info(f"Subscribed to topic: {topic}")
-
-                async for message in client.messages:
-                    logging.debug(f"Received message on topic '{message.topic}': {message.payload.decode()}")
-                    function(SnooData.from_json(message.payload.decode()))
-
-        except aiomqtt.MqttError as e:
-            logging.error(f"MQTT connection for {device.serialNumber} failed: {e}")
-        except Exception as e:
-            logging.error(f"MQTT connection for {device.serialNumber} failed with an unexpected error: {e}")
-        finally:
-            # When the connection is lost, remove the client from the map.
-            logging.info(f"MQTT connection closed for {device.serialNumber}.")
-            async with self._client_cond:
-                if device.serialNumber in self._client_map:
-                    del self._client_map[device.serialNumber]
+# {'timeout_min': 15, 'state': 'on', 'command': 'set_sticky_white_noise', 'ts': 1735237447471}
+# {'command': 'start_snoo', 'ts': 1735237642640}
+# {'command': 'go_to_state', 'ts': 1735237671846, 'state': 'LEVEL1', 'hold': 'off'}
+# {'state': 'ONLINE', 'hold': 'off', 'ts': 1735239968419, 'command': 'go_to_state'}
